/*
 * This file is a part of morm
 * and is distributed under the 2-Clause BSD licence.
 * Author: Tomasz Sowa <t.sowa@ttmath.org>
 */

/*
 * Copyright (c) 2018-2019, Tomasz Sowa
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "dbexpression.h"


namespace morm
{


DbExpression::DbExpression()
{
	output_type = 0;
}


DbExpression::~DbExpression()
{
}


void DbExpression::set_output_type(int output_type)
{
	this->output_type = output_type;
}


int DbExpression::get_output_type()
{
	return output_type;
}


bool DbExpression::can_field_be_generated(bool insertable, bool updatable, bool is_primary_key)
{
	if( output_type == MORM_OUTPUT_TYPE_DB_INSERT )
	{
		return insertable;
	}
	else
	if(	output_type == MORM_OUTPUT_TYPE_DB_UPDATE )
	{
		return updatable;
	}
	else
	if( output_type == MORM_OUTPUT_TYPE_DB_PRIMARY_KEY || output_type == MORM_OUTPUT_TYPE_JOIN_TABLES )
	{
		return is_primary_key;
	}
	else
	{
		return true;
	}
}



void DbExpression::field_before()
{
	BaseExpression::field_before();

	if( !is_first_field )
	{
		if( output_type == MORM_OUTPUT_TYPE_DB_INSERT ||
			output_type == MORM_OUTPUT_TYPE_DB_UPDATE ||
			output_type == MORM_OUTPUT_TYPE_SELECT_COLUMNS )
		{
			(*out_stream) << ",";
		}
		else
		if( output_type == MORM_OUTPUT_TYPE_DB_PRIMARY_KEY )
		{
			(*out_stream) << " AND ";
		}
		else
		if( output_type == MORM_OUTPUT_TYPE_WHERE_EQ ||
			output_type == MORM_OUTPUT_TYPE_WHERE_GE ||
			output_type == MORM_OUTPUT_TYPE_WHERE_GT ||
			output_type == MORM_OUTPUT_TYPE_WHERE_LE ||
			output_type == MORM_OUTPUT_TYPE_WHERE_LT ||
			output_type == MORM_OUTPUT_TYPE_WHERE_IN )
		{
			int conjunction = MORM_CONJUNCTION_AND;

			if( !conjunctions.empty() )
			{
				conjunction = conjunctions.back();
			}

			if( conjunction == MORM_CONJUNCTION_AND )
			{
				(*out_stream) << " AND ";
			}
			else
			if( conjunction == MORM_CONJUNCTION_OR )
			{
				(*out_stream) << " OR ";
			}
		}
	}
}

void DbExpression::put_name_value_separator()
{
	if( output_type == MORM_OUTPUT_TYPE_DB_PRIMARY_KEY ||
		output_type == MORM_OUTPUT_TYPE_WHERE_EQ ||
		output_type == MORM_OUTPUT_TYPE_DB_UPDATE)
	{
		(*out_stream) << " = ";
	}
	else
	if( output_type == MORM_OUTPUT_TYPE_WHERE_LT )
	{
		(*out_stream) << " < ";
	}
	else
	if( output_type == MORM_OUTPUT_TYPE_WHERE_GT )
	{
		(*out_stream) << " > ";
	}
	else
	if( output_type == MORM_OUTPUT_TYPE_WHERE_LE )
	{
		(*out_stream) << " <= ";
	}
	else
	if( output_type == MORM_OUTPUT_TYPE_WHERE_GE )
	{
		(*out_stream) << " >= ";
	}
	else
	if( output_type == MORM_OUTPUT_TYPE_WHERE_IN )
	{
		(*out_stream) << " in ";
	}
}



void DbExpression::before_field_value_string()
{
//	if( output_type == MORM_OUTPUT_TYPE_SELECT_COLUMNS ||
//		output_type == MORM_OUTPUT_TYPE_DB_INSERT ||
//		output_type == MORM_OUTPUT_TYPE_DB_UPDATE ||
//		output_type == MORM_OUTPUT_TYPE_DB_PRIMARY_KEY )
	{
		(*out_stream) << "'";
	}
}

void DbExpression::after_field_value_string()
{
//	if( output_type == MORM_OUTPUT_TYPE_SELECT_COLUMNS ||
//		output_type == MORM_OUTPUT_TYPE_DB_INSERT ||
//		output_type == MORM_OUTPUT_TYPE_DB_UPDATE ||
//		output_type == MORM_OUTPUT_TYPE_DB_PRIMARY_KEY )
	{
		(*out_stream) << "'";
	}
}


void DbExpression::before_field_value(const std::wstring &)
{
	before_field_value_string();
}

void DbExpression::after_field_value(const std::wstring &)
{
	after_field_value_string();
}

void DbExpression::before_field_value(const std::string &)
{
	before_field_value_string();
}

void DbExpression::after_field_value(const std::string &)
{
	after_field_value_string();
}

void DbExpression::before_field_value(const wchar_t *)
{
	before_field_value_string();
}

void DbExpression::after_field_value(const wchar_t *)
{
	after_field_value_string();
}

void DbExpression::before_field_value(const char *)
{
	before_field_value_string();
}

void DbExpression::after_field_value(const char *)
{
	after_field_value_string();
}

void DbExpression::before_field_value(const PT::Date &)
{
	before_field_value_string();
}

void DbExpression::after_field_value(const PT::Date &)
{
	after_field_value_string();
}


void DbExpression::prepare_to_where_clause()
{
	work_mode = MORM_WORK_MODE_MODEL_FIELDS_VALUES;
	conjunctions.clear();
	is_first_field = true;
}


DbExpression & DbExpression::group_or(PT::TextStream & stream)
{
	out_stream = &stream;
	field_before();
	(*out_stream) << "(";
	is_first_field = true;
	conjunctions.push_back(MORM_CONJUNCTION_OR);
	out_stream = nullptr;

	return *this;
}

DbExpression & DbExpression::group_and(PT::TextStream & stream)
{
	out_stream = &stream;
	field_before();
	(*out_stream) << "(";
	is_first_field = true;
	conjunctions.push_back(MORM_CONJUNCTION_AND);
	out_stream = nullptr;

	return *this;
}

DbExpression & DbExpression::group_end(PT::TextStream & stream)
{
	out_stream = &stream;

	if( !conjunctions.empty() )
	{
		conjunctions.pop_back();
		(*out_stream) << ")";
		field_after();
	}

	out_stream = nullptr;
	return *this;
}


DbExpression & DbExpression::page(PT::TextStream & stream, size_t page_number, size_t page_size)
{
	stream << " limit " << page_number << "," << page_size << " ";
	return *this;
}

}
