/*
 * This file is a part of morm
 * and is distributed under the 2-Clause BSD licence.
 * Author: Tomasz Sowa <t.sowa@ttmath.org>
 */

/*
 * Copyright (c) 2018-2019, Tomasz Sowa
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef headerfile_morm_finder
#define headerfile_morm_finder

#include <vector>
#include <list>
#include <set>
#include "model.h"
#include "utf8/utf8.h"
#include "textstream/textstream.h"
#include "dbconnector.h"
#include "modelconnector.h"
#include "cursor.h"


namespace morm
{

template<typename ModelClass>
class Finder
{

public:


	Finder()
	{
		model_connector = nullptr;
		out_stream = nullptr;
		db_expression = nullptr;
		was_query_error = false;
		model_data = nullptr;
	}


	Finder(PT::TextStream & out_stream)
	{
		this->out_stream = &out_stream;
		this->model_connector = nullptr;
		db_expression = nullptr;
		was_query_error = false;
		model_data = nullptr;
	}

	Finder(ModelConnector & model_connector)
	{
		this->model_connector = &model_connector;
		out_stream = nullptr;
		db_expression = nullptr;
		was_query_error = false;
		model_data = nullptr;
		set_out_stream();
	}

	Finder(ModelConnector * model_connector)
	{
		this->model_connector = model_connector;
		out_stream = nullptr;
		db_expression = nullptr;
		was_query_error = false;
		model_data = nullptr;
		set_out_stream();
	}

	Finder(PT::TextStream & out_stream, ModelConnector & model_connector)
	{
		this->out_stream = &out_stream;
		this->model_connector = &model_connector;
		db_expression = nullptr;
		was_query_error = false;
		model_data = nullptr;
	}

	Finder(PT::TextStream & out_stream, ModelConnector * model_connector)
	{
		this->out_stream = &out_stream;
		this->model_connector = model_connector;
		db_expression = nullptr;
		was_query_error = false;
		model_data = nullptr;
	}




protected:

	void set_out_stream()
	{
		if( model_connector )
		{
			this->out_stream = model_connector->get_stream();
		}
		else
		{
			this->out_stream = nullptr;
		}
	}


public:

	Finder<ModelClass> & set_out_stream(PT::TextStream * out_stream)
	{
		this->out_stream = out_stream;
		return *this;
	}

	Finder<ModelClass> & set_model_data(ModelData * model_data)
	{
		this->model_data = model_data;
		return *this;
	}

	Finder<ModelClass> & set_model_data(ModelData & model_data)
	{
		this->model_data = &model_data;
		return *this;
	}


	bool was_error()
	{
		return was_query_error;
	}


	std::wstring get_error_msg()
	{
		return last_query_error;
	}


	Finder<ModelClass> & prepare_to_select()
	{
		was_query_error = false;
		last_query_error.clear();
		has_autogenerated_select = false;

		if( model_connector )
		{
			if( out_stream )
			{
				set_db_expression();
				out_stream->clear();
			}
			else
			{
				was_query_error = true;
				last_query_error = L"out stream object is required";
			}
		}
		else
		{
			was_query_error = true;
			last_query_error = L"model connector object is required";
		}

		return *this;
	}


	Finder<ModelClass> & select(ModelConnector & model_connector)
	{
		this->model_connector = &model_connector;
		set_out_stream();

		return select();
	}

	Finder<ModelClass> & select(ModelConnector * model_connector)
	{
		this->model_connector = model_connector;
		set_out_stream();

		return select();
	}

	Finder<ModelClass> & select(PT::TextStream & out_stream, ModelConnector & model_connector)
	{
		this->out_stream = &out_stream;
		this->model_connector = &model_connector;
		set_out_stream();

		return select();
	}

	Finder<ModelClass> & select(PT::TextStream & out_stream, ModelConnector * model_connector)
	{
		this->out_stream = &out_stream;
		this->model_connector = model_connector;
		set_out_stream();

		return select();
	}


	Finder<ModelClass> & select(bool call_prepare = true)
	{
		if( call_prepare )
		{
			prepare_to_select();
		}

		model.set_connector(model_connector);

		finder_helper.clear();
		model_env.clear();

		model.model_env = &model_env;
		model.model_env->model_data = model_data;
		model.model_env->finder_helper = &finder_helper;
		model.prepare_to_select();

		has_autogenerated_select = true;

		if( model_connector && out_stream )
		{
			model.table_name(model.model_env->table_name);
			model.table_name_for_join_as(model.model_env->table_name_simple);
			model.model_env->table_index = finder_helper.add_join_table(model.model_env->table_name_simple);

			(*out_stream) << "SELECT ";
			model.generate_select_columns(*out_stream);
			(*out_stream) << " FROM " << model.model_env->table_name << " AS ";
			(*out_stream) << model.model_env->table_name_simple;
			(*out_stream) << " ";
			(*out_stream) << finder_helper.morm_finder_join_tables;
		}

		return *this;
	}


	Finder<ModelClass> & select(ModelData * model_data, bool call_prepare = true)
	{
		if( call_prepare )
		{
			prepare_to_select();
		}

		this->model_data = model_data;
		return select(false);
	}


	Finder<ModelClass> & select(ModelData & model_data, bool call_prepare = true)
	{
		if( call_prepare )
		{
			prepare_to_select();
		}

		this->model_data = &model_data;
		return select(false);
	}


	Finder<ModelClass> & where()
	{
		if( out_stream && db_expression )
		{
			(*out_stream) << " WHERE ";
			db_expression->prepare_to_where_clause();
		}

		return *this;
	}


	Finder<ModelClass> & group_or()
	{
		if( db_expression && out_stream )
		{
			db_expression->group_or(*out_stream);
		}

		return *this;
	}


	Finder<ModelClass> & group_and()
	{
		if( db_expression && out_stream )
		{
			db_expression->group_and(*out_stream);
		}

		return *this;
	}


	Finder<ModelClass> & group_end()
	{
		if( db_expression && out_stream )
		{
			db_expression->group_end(*out_stream);
		}

		return *this;
	}




	template<typename FieldValue>
	Finder<ModelClass> & eq(const wchar_t * field_name, const FieldValue & field_value)
	{
		if( db_expression )
		{
			db_expression->set_output_type(MORM_OUTPUT_TYPE_WHERE_EQ);

			bool put_table_prefix = true;

			// IMPROVE ME
			// add check if there is a dot char in field_name?
			// or add a second eq() method with an additional parameter -- table name?

			db_expression->field_to_stream(*out_stream, field_name, field_value, false, false, false, put_table_prefix, &model_env);
		}

		return *this;
	}


	template<typename FieldValue>
	Finder<ModelClass> & lt(const wchar_t * field_name, const FieldValue & field_value)
	{
		if( db_expression )
		{
			db_expression->set_output_type(MORM_OUTPUT_TYPE_WHERE_LT);
			db_expression->field_to_stream(*out_stream, field_name, field_value, false, false, false, true, &model_env);
		}

		return *this;
	}


	template<typename FieldValue>
	Finder<ModelClass> & gt(const wchar_t * field_name, const FieldValue & field_value)
	{
		if( db_expression )
		{
			db_expression->set_output_type(MORM_OUTPUT_TYPE_WHERE_GT);
			db_expression->field_to_stream(*out_stream, field_name, field_value, false, false, false, true, &model_env);
		}

		return *this;
	}


	template<typename FieldValue>
	Finder<ModelClass> & le(const wchar_t * field_name, const FieldValue & field_value)
	{
		if( db_expression )
		{
			db_expression->set_output_type(MORM_OUTPUT_TYPE_WHERE_LE);
			db_expression->field_to_stream(*out_stream, field_name, field_value, false, false, false, true, &model_env);
		}

		return *this;
	}


	template<typename FieldValue>
	Finder<ModelClass> & ge(const wchar_t * field_name, const FieldValue & field_value)
	{
		if( db_expression )
		{
			db_expression->set_output_type(MORM_OUTPUT_TYPE_WHERE_GE);
			db_expression->field_to_stream(*out_stream, field_name, field_value, false, false, false, true, &model_env);
		}

		return *this;
	}



	template<typename FieldValue>
	Finder<ModelClass> & in(const wchar_t * field_name, const std::set<FieldValue> & container)
	{
		if( db_expression )
		{
			db_expression->set_output_type(MORM_OUTPUT_TYPE_WHERE_IN);
			db_expression->field_in(*out_stream, field_name, container);
		}

		return *this;
	}

	template<typename FieldValue>
	Finder<ModelClass> & in(const wchar_t * field_name, const std::list<FieldValue> & container)
	{
		if( db_expression )
		{
			db_expression->set_output_type(MORM_OUTPUT_TYPE_WHERE_IN);
			db_expression->field_in(*out_stream, field_name, container);
		}

		return *this;
	}

	template<typename FieldValue>
	Finder<ModelClass> & in(const wchar_t * field_name, const std::vector<FieldValue> & container)
	{
		if( db_expression )
		{
			db_expression->set_output_type(MORM_OUTPUT_TYPE_WHERE_IN);
			db_expression->field_in(*out_stream, field_name, container);
		}

		return *this;
	}

	Finder<ModelClass> & page(size_t page_number, size_t page_size)
	{
		if( out_stream && db_expression )
		{
			db_expression->page(*out_stream, page_number, page_size);
		}

		return *this;
	}


	Finder<ModelClass> & raw(const char * sql, bool add_spaces = true)
	{
		if( out_stream )
		{
			if( add_spaces )
				(*out_stream) << ' ';

			(*out_stream) << sql;

			if( add_spaces )
				(*out_stream) << ' ';
		}

		return *this;
	}


	Finder<ModelClass> & raw(PT::TextStream & sql, bool add_spaces = true)
	{
		if( out_stream )
		{
			if( add_spaces )
				(*out_stream) << ' ';

			(*out_stream) << sql;

			if( add_spaces )
				(*out_stream) << ' ';
		}

		return *this;
	}


	Finder<ModelClass> & raw(long val, bool add_spaces = true)
	{
		if( out_stream )
		{
			if( add_spaces )
				(*out_stream) << ' ';

			(*out_stream) << val;

			if( add_spaces )
				(*out_stream) << ' ';
		}

		return *this;
	}


	Cursor<ModelClass> get_cursor()
	{
		Cursor<ModelClass> cursor;
		cursor.set_model_data(model_data);
		cursor.set_has_autogenerated_select(has_autogenerated_select);

		if( model_connector && out_stream )
		{
			cursor.set_model_connector(model_connector);
			DbConnector * db_connector = model_connector->get_db_connector();

			if( db_connector )
			{
				QueryResult * query_result = db_connector->create_query_result();
				cursor.set_query_result(query_result);
				query_result->references_count = 1;

				bool status = db_connector->query_select(*out_stream, *query_result);
				cursor.set_select_status(status);
			}
		}

		return cursor;
	}


	bool get(ModelClass & result)
	{
		Cursor<ModelClass> cursor = get_cursor();
		return cursor.get(result);
	}


	ModelClass get()
	{
		ModelClass model;

		get(model);
		return model;
	}


	bool get_list(std::list<ModelClass> & result, bool clear_list = true)
	{
		Cursor<ModelClass> cursor = get_cursor();
		return cursor.get_list(result, clear_list);
	}


	std::list<ModelClass> get_list()
	{
		std::list<ModelClass> result;

		get_list(result, false);
		return result;
	}




private:


	PT::TextStream * out_stream;
	ModelConnector * model_connector;
	DbExpression * db_expression;
	ModelClass model;
	bool was_query_error;
	std::wstring last_query_error;
	ModelEnv model_env;
	FinderHelper finder_helper;
	ModelData * model_data;
	bool has_autogenerated_select;


	void set_db_expression()
	{
		if( model_connector )
		{
			DbConnector * db_connector = model_connector->get_db_connector();

			if( db_connector )
			{
				db_expression = db_connector->get_expression();
				db_expression->allow_to_use_prefix(true);
			}
		}
	}


};

} // namespace

#endif
